////////////////////////////////////////////////////////////////////////////
//
// Disparity estimator
// by Hansung Kim
//
// ver 1.2 - OpenCV interface
// ver 1.3 - free input size (no need to be multiply of block size), CCW rotation for Spheron
// ver 1.4 - Canny edge detector is changed to using OpenCV to save memory
// ver 1.5 - Linux compatible version 
// ver 1.6 - RGB color based matching
// ver 1.7 - Hierarchical regularization Downsampling must be power of 2^(hierarchy)
// ver 2.0 - (1) NCC/MAE option added for matching 
//			 (2) Checking the second best matching option added for bi-directional matching
//			 (3) Loading initial disparity from file added for manual initialization
// ver 2.1 - deal with large disparity bigger than 255 pixel. (floating point scale for disparity map
// ver 3.0 - Name changed to Single Disp. Only for oneway disparity estimation. accept dense disp as input
// ver 3.1 - Error refinement in DenseDisparityRegionDividing
// ver 3.2 - Hierarchical dense disparity estimation from full resolution (debugged of ringing artefact)
// ver 3.4 - skip black lines in estimation / bug fixed in cvSobel 
// ver 3.7 - command line input
// ver 3.8 - Output of intermediate dense disparity with occlusion
// ver 3.9 - Get Iteration number for regularization
// ver 4.0 - With OpenCV 3.1. Less options 
///////////////////////////////////////////////////////////////////////////

#include <assert.h>
#include <stdio.h>
#include <stdlib.h>
#include <opencv/cv.h>
#include <opencv/highgui.h>

#include "param.h"
#include "memory.h"
#include "util.h"
#include "disparity.h"

int main(int argc, char* argv[])
{

	//For commnad line input /////////////////////////////////////////////
	if (argc < 4)
	{
		printf("Syntax is: %s Left(Top)_Image Right(Bottom)_Image Starting_point\n", argv[0]);
		//printf("	Matching method: 0=MAE, 1=NCC\n");
		printf("	Starting point: 0=Full, 1= From initial, 2=From Dense\n");
		printf("	matching Method: MAE, Search range: -120. No regularization. use V3.9 if you want different options\n");
		return (0);
	}

	//matching_method = atoi(argv[3]);
	matching_method = 0;
	initial_disp_from_file = atoi(argv[3]);
	//iteration = atoi(argv[5]);
	iteration = 0;

	////////////////////////////////////////////////////////////////////

	int				xsize_d, ysize_d;		// down sampled image size
	float			scale;					//Scale factor for display

											//Disparity estimation
	float			**Disparity;
	int				**DisparityInt;

	FILE			*fr;
	char			filename[200];

	int i, j;
	int temp;
	char path[300];
	CvSize	Size_offset;
	CvSize	rotated_size;

	IplImage			*left_org, *right_org;	//color
	IplImage			*left_org_grey, *right_org_grey;


	//Load data for parameter file input
	/*
	sprintf(filename, "%s%s", BASE_PATH, LEFT_IMAGE);
	if((left_org=cvLoadImage(filename)) == 0)
	{
	printf("Cannot open the first image - %s\n",filename);
	exit(0);
	}

	sprintf(filename, "%s%s", BASE_PATH, RIGHT_IMAGE);
	if((right_org=cvLoadImage(filename)) == 0)
	{
	printf("Cannot open the second image - %s\n",filename);
	exit(0);
	}
	*/


	//Load data for command line input  ////////////////////////////////////////////

	sprintf(filename, "%s%s", BASE_PATH, argv[1]);
	if ((left_org = cvLoadImage(filename)) == 0)
	{
		printf("Cannot open the first image - %s\n", filename);
		exit(0);
	}

	sprintf(filename, "%s%s", BASE_PATH, argv[2]);
	if ((right_org = cvLoadImage(filename)) == 0)
	{
		printf("Cannot open the second image - %s\n", filename);
		exit(0);
	}
	////////////////////////////////////////////////////////////////////////////////

	Imagesize = cvGetSize(left_org);
	printf("Image size: %d %d \n\n", Imagesize.width, Imagesize.height);

	Imagesize_d.width = Imagesize.width / down;
	Imagesize_d.height = Imagesize.height / down;

	int down_ratio = pow(2.0, hierarchy);

	if (Imagesize_d.width % (Block*down_ratio) == 0)
		Size_offset.width = 0;
	else
		Size_offset.width = Block*down_ratio - (Imagesize_d.width % (Block*down_ratio));

	if (Imagesize_d.height % (Block*down_ratio) == 0)
		Size_offset.height = 0;
	else
		Size_offset.height = Block*down_ratio - (Imagesize_d.height % (Block*down_ratio));

	Imagesize_ext.width = Imagesize_d.width + Size_offset.width;
	Imagesize_ext.height = Imagesize_d.height + Size_offset.height;

	printf("Extended image size = %d %d \n\n", Imagesize_ext.width, Imagesize_ext.height);

	scale = 255.0 / (Max_Range - Min_Range);

	if (scale >= 1.0)
		scale = (int)scale;
	else if (scale >= 0.5)
		scale = 0.5;
	else if (scale >= 0.25)
		scale = 0.25;
	else
		scale = 1.0;

	printf("Disparity scale = %f, scale offset = %d \n\n", scale, (int)(Min_Range));

	//Splitting to RGB channels (left image)
	IplImage*		left_r = cvCreateImage(Imagesize, 8, 1);
	IplImage*		left_g = cvCreateImage(Imagesize, 8, 1);
	IplImage*		left_b = cvCreateImage(Imagesize, 8, 1);
	left_org_grey = cvCreateImage(Imagesize, 8, 1);

	cvSplit(left_org, left_b, left_g, left_r, NULL);
	cvCvtColor(left_org, left_org_grey, CV_BGR2GRAY);

	cvReleaseImage(&left_org);

	cvSmooth(left_b, left_b, CV_GAUSSIAN);
	cvSmooth(left_g, left_g, CV_GAUSSIAN);
	cvSmooth(left_r, left_r, CV_GAUSSIAN);


	//Downsampling
	IplImage*		left_b_d = cvCreateImage(Imagesize_d, 8, 1);
	IplImage*		left_g_d = cvCreateImage(Imagesize_d, 8, 1);
	IplImage*		left_r_d = cvCreateImage(Imagesize_d, 8, 1);

	cvResize(left_b, left_b_d);
	cvResize(left_g, left_g_d);
	cvResize(left_r, left_r_d);

	cvReleaseImage(&left_b);
	cvReleaseImage(&left_g);
	cvReleaseImage(&left_r);

	//make image to multiply of Block size
	IplImage*		left_r_ext = cvCreateImage(Imagesize_ext, 8, 1);
	IplImage*		left_g_ext = cvCreateImage(Imagesize_ext, 8, 1);
	IplImage*		left_b_ext = cvCreateImage(Imagesize_ext, 8, 1);

	cvSetZero(left_b_ext);
	cvSetZero(left_g_ext);
	cvSetZero(left_r_ext);

	for (i = 0; i<Imagesize_d.height; i++)
		for (j = 0; j<Imagesize_d.width; j++)
		{
			cvSetReal2D(left_b_ext, i, j, cvGetReal2D(left_b_d, i, j));
			cvSetReal2D(left_g_ext, i, j, cvGetReal2D(left_g_d, i, j));
			cvSetReal2D(left_r_ext, i, j, cvGetReal2D(left_r_d, i, j));
		}

	cvReleaseImage(&left_b_d);
	cvReleaseImage(&left_g_d);
	cvReleaseImage(&left_r_d);

	//Data transfer
	UnsignedRGB		**left_d;
	if (rotation == 1)
	{
		ysize_d = Imagesize_ext.width;
		xsize_d = Imagesize_ext.height;
		left_d = Allo2RGB(ysize_d, xsize_d);

		for (i = 0; i<Imagesize_ext.height; i++)
			for (j = 0; j<Imagesize_ext.width; j++)
			{
				left_d[Imagesize_ext.width - j - 1][i].b = cvGetReal2D(left_b_ext, i, j);
				left_d[Imagesize_ext.width - j - 1][i].g = cvGetReal2D(left_g_ext, i, j);
				left_d[Imagesize_ext.width - j - 1][i].r = cvGetReal2D(left_r_ext, i, j);
			}

		rotated_size = cvSize(Imagesize.height, Imagesize.width);
		IplImage*	temp_img = cvCreateImage(rotated_size, 8, 1);
		for (i = 0; i<Imagesize.height; i++)
			for (j = 0; j<Imagesize.width; j++)
				cvSetReal2D(temp_img, Imagesize.width - j - 1, i, cvGetReal2D(left_org_grey, i, j));
		cvReleaseImage(&left_org_grey);
		left_org_grey = cvCloneImage(temp_img);
		cvReleaseImage(&temp_img);
	}
	else
	{
		xsize_d = Imagesize_ext.width;
		ysize_d = Imagesize_ext.height;
		left_d = Allo2RGB(ysize_d, xsize_d);

		for (i = 0; i<Imagesize_ext.height; i++)
			for (j = 0; j<Imagesize_ext.width; j++)
			{
				left_d[i][j].r = cvGetReal2D(left_r_ext, i, j);
				left_d[i][j].g = cvGetReal2D(left_g_ext, i, j);
				left_d[i][j].b = cvGetReal2D(left_b_ext, i, j);
			}
	}
	cvReleaseImage(&left_r_ext);
	cvReleaseImage(&left_g_ext);
	cvReleaseImage(&left_b_ext);

	///////Repeat for right image
	//Splitting to RGB channels (right image)
	IplImage*		right_r = cvCreateImage(Imagesize, 8, 1);
	IplImage*		right_g = cvCreateImage(Imagesize, 8, 1);
	IplImage*		right_b = cvCreateImage(Imagesize, 8, 1);
	right_org_grey = cvCreateImage(Imagesize, 8, 1);

	cvSplit(right_org, right_b, right_g, right_r, NULL);
	cvCvtColor(right_org, right_org_grey, CV_BGR2GRAY);

	cvReleaseImage(&right_org);

	cvSmooth(right_b, right_b, CV_GAUSSIAN);
	cvSmooth(right_g, right_g, CV_GAUSSIAN);
	cvSmooth(right_r, right_r, CV_GAUSSIAN);

	//Downsampling
	IplImage*		right_b_d = cvCreateImage(Imagesize_d, 8, 1);
	IplImage*		right_g_d = cvCreateImage(Imagesize_d, 8, 1);
	IplImage*		right_r_d = cvCreateImage(Imagesize_d, 8, 1);

	cvResize(right_b, right_b_d);
	cvResize(right_g, right_g_d);
	cvResize(right_r, right_r_d);

	cvReleaseImage(&right_b);
	cvReleaseImage(&right_g);
	cvReleaseImage(&right_r);

	//make image to multiply of Block size
	IplImage*		right_r_ext = cvCreateImage(Imagesize_ext, 8, 1);
	IplImage*		right_g_ext = cvCreateImage(Imagesize_ext, 8, 1);
	IplImage*		right_b_ext = cvCreateImage(Imagesize_ext, 8, 1);

	cvSetZero(right_b_ext);
	cvSetZero(right_g_ext);
	cvSetZero(right_r_ext);

	for (i = 0; i<Imagesize_d.height; i++)
		for (j = 0; j<Imagesize_d.width; j++)
		{
			cvSetReal2D(right_b_ext, i, j, cvGetReal2D(right_b_d, i, j));
			cvSetReal2D(right_g_ext, i, j, cvGetReal2D(right_g_d, i, j));
			cvSetReal2D(right_r_ext, i, j, cvGetReal2D(right_r_d, i, j));
		}

	cvReleaseImage(&right_b_d);
	cvReleaseImage(&right_g_d);
	cvReleaseImage(&right_r_d);

	//Data transfer
	UnsignedRGB		**right_d;
	if (rotation == 1)
	{
		right_d = Allo2RGB(ysize_d, xsize_d);

		for (i = 0; i<Imagesize_ext.height; i++)
			for (j = 0; j<Imagesize_ext.width; j++)
			{
				right_d[Imagesize_ext.width - j - 1][i].b = cvGetReal2D(right_b_ext, i, j);
				right_d[Imagesize_ext.width - j - 1][i].g = cvGetReal2D(right_g_ext, i, j);

				right_d[Imagesize_ext.width - j - 1][i].r = cvGetReal2D(right_r_ext, i, j);
			}

		IplImage*	temp_img2 = cvCreateImage(rotated_size, 8, 1);
		for (i = 0; i<Imagesize.height; i++)
			for (j = 0; j<Imagesize.width; j++)
				cvSetReal2D(temp_img2, Imagesize.width - j - 1, i, cvGetReal2D(right_org_grey, i, j));
		cvReleaseImage(&right_org_grey);
		right_org_grey = cvCloneImage(temp_img2);
		cvReleaseImage(&temp_img2);

	}
	else
	{
		right_d = Allo2RGB(ysize_d, xsize_d);

		for (i = 0; i<Imagesize_ext.height; i++)
			for (j = 0; j<Imagesize_ext.width; j++)
			{
				right_d[i][j].r = cvGetReal2D(right_r_ext, i, j);
				right_d[i][j].g = cvGetReal2D(right_g_ext, i, j);
				right_d[i][j].b = cvGetReal2D(right_b_ext, i, j);
			}
	}
	cvReleaseImage(&right_r_ext);
	cvReleaseImage(&right_g_ext);
	cvReleaseImage(&right_b_ext);


	//Main function - Disparity estimation 
	Disparity = Allo2Float(ysize_d, xsize_d);

	sprintf(path, "%s", BASE_PATH);

	DisparityWithRegularization(left_d, right_d, left_org_grey, right_org_grey, Disparity, xsize_d, ysize_d,
		hierarchy, Block, Min_Range, Max_Range, scale, Epsilon,
		Region_Th, beta, lambda, iteration, canny_Th, path, initial_disp_from_file);

	cvReleaseImage(&left_org_grey);
	cvReleaseImage(&right_org_grey);

	//save results
	CvSize output_size = cvSize(Imagesize_d.width, Imagesize_d.height);
	CvSize result_size;

	IplImage*		result1 = cvCreateImage(output_size, 8, 1);
	IplImage*		result2 = cvCreateImage(output_size, 8, 3);
	IplImage*		result_r = cvCreateImage(output_size, 8, 1);
	IplImage*		result_g = cvCreateImage(output_size, 8, 1);
	IplImage*		result_b = cvCreateImage(output_size, 8, 1);

	int				temp2;
	int				temp3, temp4, temp5;

	if (rotation == 1)
	{
		result_size = cvSize(ysize_d, xsize_d);

		for (i = 0; i<output_size.height; i++)
			for (j = 0; j<output_size.width; j++)
			{
				temp = (int)(scale * (Disparity[result_size.width - j - 1][i] - Min_Range) + 0.5);
				temp2 = (int)(1000 * (Disparity[result_size.width - j - 1][i] - Min_Range) + 0.5);

				if (temp<0)
					cvSetReal2D(result1, i, j, 0);
				else if (temp> 255)
					cvSetReal2D(result1, i, j, 255);
				else
					cvSetReal2D(result1, i, j, temp);

				if (temp2<0)
				{
					cvSetReal2D(result_r, i, j, 0);
					cvSetReal2D(result_g, i, j, 0);
					cvSetReal2D(result_b, i, j, 0);
				}
				else
				{
					temp3 = temp2 / 256;
					temp4 = temp2 % 256;
					cvSetReal2D(result_r, i, j, temp4);
					temp4 = temp3 / 256;
					temp5 = temp3 % 256;
					cvSetReal2D(result_g, i, j, temp5);
					cvSetReal2D(result_b, i, j, temp4);
				}

			}
	}
	else
	{
		result_size = cvSize(xsize_d, ysize_d);

		for (i = 0; i<output_size.height; i++)
			for (j = 0; j<output_size.width; j++)
			{
				temp = (int)(scale * (Disparity[i][j] - Min_Range) + 0.5);
				temp2 = (int)(1000 * (Disparity[i][j] - Min_Range) + 0.5);

				if (temp<0)
					cvSetReal2D(result1, i, j, 0);
				else if (temp> 255)
					cvSetReal2D(result1, i, j, 255);
				else
					cvSetReal2D(result1, i, j, temp);

				if (temp2<0)
				{
					cvSetReal2D(result_r, i, j, 0);
					cvSetReal2D(result_g, i, j, 0);
					cvSetReal2D(result_b, i, j, 0);
				}
				else
				{
					temp3 = temp2 / 256;
					temp4 = temp2 % 256;
					cvSetReal2D(result_r, i, j, temp4);
					temp4 = temp3 / 256;
					temp5 = temp3 % 256;
					cvSetReal2D(result_g, i, j, temp5);
					cvSetReal2D(result_b, i, j, temp4);
				}
			}
	}

	cvMerge(result_b, result_g, result_r, NULL, result2);

	sprintf(filename, "%s//05Disparity-g_%d_%d.png", BASE_PATH, (int)scale, (int)(Min_Range));
	cvSaveImage(filename, result1);
	sprintf(filename, "%s//05Disparity-rgb_1000_%d.png", BASE_PATH, (int)(Min_Range));
	cvSaveImage(filename, result2);


	/*
	char *title = "result";
	cvNamedWindow( title, CV_WINDOW_AUTOSIZE );
	cvShowImage( title, result );
	cvWaitKey(0);
	cvDestroyWindow( title );
	*/

	Deallo2RGB(left_d, ysize_d);
	Deallo2RGB(right_d, ysize_d);
	Deallo2Float(Disparity, ysize_d);
	//Deallo2Int(DisparityInt,ysize_d);

	cvReleaseImage(&result1);
	cvReleaseImage(&result2);
	cvReleaseImage(&result_r);
	cvReleaseImage(&result_g);
	cvReleaseImage(&result_b);

	return 0;
}
