/////////////////////////////////////////////////////////
//
//  Room Model Reconstruction for S3A Project
//  By Hansung Kim (coolhs99@gmail.com)
//
// RoomModelSegNetBlockReconstruction V1.0
//		- Modified from RoomModelBlockReconstruction V1.0 which was 
//			- Simple Reconstruction from Integer disparity with occlusion
//			- facade alignment with only one face
//			- Optionally include Object recognition results
//			- Optionally Segmentation with Graph cut + Object recognition
//			- User interaction directly assigne object number
//			- Add rectangle
//			- User correction
//			- Coordinate changed from left/back/up in Plane_REconstruction V3.3 to Right/Front/Up
//
//		- Changed to OpenCV V3.1 (based on RoomModelCorrectionTool-V1.1)
//		- Removed Graph Cut and Eigen-Faug. object recognition
//		- Work with SegNet result 
//		- Input should be: shifted_t.png, shifted_b.png, shifted-disparity.png and 13_obj_index.png
//		- Corection of texture coordinate for model with texture
//		- Erosion of regions filed in reconstruction. 
//		- Object extension to ground and walls 
//
// V1.1 - Debug room boundary reconstruction for large empty rooms
////////////////////////////////////////////////////////////////////////////////

// Coordinate: x:right y:front z:up (It is different from RGBD coordinate which is z is up and y is toward)



#include <opencv2/core.hpp>
#include <opencv2/imgcodecs.hpp>
#include <opencv2/highgui.hpp>
#include <opencv2/imgproc.hpp>
#include <iostream>
#include <time.h>

using namespace cv;
using namespace std;

#include "variable.h"
#include "IO.h"
#include "region.h"
#include "UI.h"
#include "reconstruction.h"
#include "jWrite.h"


int main(int argc, char** argv)
{
	//For commnad line input /////////////////////////////////////////////
	if (argc != 6)
	{
		cout << "Usage: " << argv[0] << " [Base Path] [baseline diatance (in m)] [UI flag] [Output name (Prefix)] [Save Option]" << endl;
		cout << "	UI flag: 0: Without User Interaction 1:With User Interaction 2: Edit pre-edited index file again 3: Use pre-edited index file\n" << endl;
		cout << "	Save Option: 0: OBJ with JASON  1: OBJ with obj colours and JASON  2: OBJ with texture 3: OFF with obj colours" << endl;
		return -1;
	}

	BASE_PATH = argv[1];
	baseline = atof(argv[2]);
	UI_flag = atoi(argv[3]);
	OUTPUT_PREFIXE = argv[4];
	OBJ_SAVE = atoi(argv[5]);


	clock_t start_time, end_time;
	start_time = clock();

	//Load input images
	img_t = imread((BASE_PATH + TOP_IMAGE));
	if (img_t.empty())
	{
		cout << "Could not open or find the image: " << TOP_IMAGE << endl;
		return -1;
	}

	img_b = imread((BASE_PATH + BOTTOM_IMAGE));
	if (img_b.empty())
	{
		cout << "Could not open or find the image: " << BOTTOM_IMAGE << endl;
		return -1;
	}

	img_disparity = imread((BASE_PATH + DISP_IMAGE), -1);
	if (img_disparity.empty())
	{
		cout << "Could not open or find the image: " << DISP_IMAGE << endl;
		return -1;
	}
	disp_ch = img_disparity.channels();
	cout << "Disparity channel: " << disp_ch << endl;

	if (UI_flag == 0 || UI_flag == 1)		//Use the original object index file
	{
		obj_index = imread((BASE_PATH + OBJ_INDEX_NAME), 0);
		if (obj_index.empty())
		{
			cout << "Could not open or find the image: " << OBJ_INDEX_NAME << endl;
			return -1;
		}
	}

	else 	//Use the pre-edited object index file
	{
		obj_index = imread((BASE_PATH + OBJ_INDEX_NAME_NEW), 0);
		if (obj_index.empty())
		{
			cout << "Could not open or find the image: " << OBJ_INDEX_NAME_NEW << endl;
			return -1;
		}
	}



	end_time = clock();
	cout << "Input data loading: " << ((double)(end_time - start_time) / CLOCKS_PER_SEC) << "secs" << endl;

	img_size = img_t.size();
	unit_h = 1.0 / (img_size.height);
	unit_w = 2.0 / (img_size.width);

	//Disparity conversion
	disparity = Mat::zeros(img_disparity.size(), CV_32FC1);
	if (disp_ch == 3)
	{
		disp_scale = 1000;
		for (int i = 0; i < img_t.rows; i++)
			for (int j = 0; j < img_t.cols; j++)
				disparity.at<float>(i, j) = (img_disparity.at<Vec3b>(i, j)[0] * 65536 + img_disparity.at<Vec3b>(i, j)[1] * 256 + img_disparity.at<Vec3b>(i, j)[2]);
	}
	else
	{
		disp_scale = 2;
		for (int i = 0; i < img_t.rows; i++)
			for (int j = 0; j < img_t.cols; j++)
				disparity.at<float>(i, j) = img_disparity.at<uchar>(i, j);
	}

	//Filtering of index image
	Mat filtered;
	medianBlur(obj_index, filtered, 9);
	medianBlur(filtered, obj_index, 9);

	// Conversion of index to index_colour
	obj_index_colour = img_t.clone();	//just to make a colour matrix with the same size
	index2indexColour(obj_index, obj_index_colour);

	if (UI_flag == 1 || UI_flag == 2)
	{
		//Region refinement (user interaction)
		start_time = clock();

		Mat img_resize, index_resize, index_colour_resize;
		Size reformed_size;
		float reform_factor;

		if (img_t.cols > 1800)
			reform_factor = 2;
		else
			reform_factor = 1;

		reformed_size.height = img_size.height / reform_factor;
		reformed_size.width = img_size.width / reform_factor;

		resize(img_t, img_resize, reformed_size, 0, 0, INTER_NEAREST);
		resize(obj_index, index_resize, reformed_size, 0, 0, INTER_NEAREST);
		resize(obj_index_colour, index_colour_resize, reformed_size, 0, 0, INTER_NEAREST);

		manual_index_processing(img_resize, index_resize, index_colour_resize);

		resize(index_resize, obj_index, img_size, 0,0, INTER_NEAREST);

		medianBlur(obj_index, filtered, 9);
		medianBlur(filtered, obj_index, 9);

		index2indexColour(obj_index, obj_index_colour);

		imwrite((BASE_PATH + "13_obj_index_new.png"), obj_index);
		imwrite((BASE_PATH + "13_obj_index_colour_new.png"), obj_index_colour);

		namedWindow("Colour Index", WINDOW_NORMAL);
		imshow("Colour Index", obj_index_colour);
		resizeWindow("Colour Index", 1200, 600);
		waitKey(0);
		destroyWindow("Colour Index");

		end_time = clock();
		cout << "Manual processing: " << ((double)(end_time - start_time) / CLOCKS_PER_SEC) << "secs" << endl;
	}

	////Obj Index post-processing. 
	// change picture to wall, eliminate window (to nothing)
	obj_index_post_processing(obj_index);


	//Region Indexing
	cout << "Region indexing" << endl;
	start_time = clock();

	region = Mat::zeros(obj_index.size(), CV_32SC1);
	int region_counter = region_indexing(obj_index, region);

	region_info = Mat::zeros(Size(4, region_counter+1), CV_32SC1);
	region_object_indexing( obj_index, region, region_info, region_counter + 1);

	cout << "Number of valid regions: " << region_counter << endl;

	//Mat temp_disp= obj_index.clone();
	//for (int y = 0; y<obj_index_colour.rows; y++)
	//	for (int x = 0; x<obj_index_colour.cols; x++)
	//		temp_disp.at<uchar>(Point(x, y)) = 4* region.at<int>(Point(x, y));

	//namedWindow("region", WINDOW_NORMAL);
	//imshow("region", temp_disp);
	//resizeWindow("region", 1200, 600);
	//waitKey(0);
	//destroyWindow("region");

	end_time = clock();
	cout << "Region indexing: " << ((double)(end_time - start_time) / CLOCKS_PER_SEC) << "secs" << endl;

	
	
	////cuboid info generation for valid regions
	start_time = clock();

	max_vertice = img_size.height * img_size.width;
	v_2D = new CvPoint2D32f[max_vertice];
	vertex = new CvPoint3D32f[max_vertice];
	cube_info = new float*[MAX_OBJECTS];
	for (int i = 0; i<MAX_OBJECTS; i++)
	{
		cube_info[i] = new float[9];	//[0]type (0:Invalid 1:Cuboid 2:Room bounding box) [1]Min x [2]Max_x [3]Min y [4]max y [5]min z [6] max z [7] object index [8]Material index
	}
	new_cube_info = new float[9];

	num_cube = 0;

	cout << "Boundary estimation" << endl;
	room_boundary_reconstruction(obj_index, disparity);

	cube_info[0][0] = 2; cube_info[0][1] = room_min.x;  cube_info[0][2] = room_max.x;
	cube_info[0][3] = room_min.y;  cube_info[0][4] = room_max.y; cube_info[0][5] = room_min.z;  cube_info[0][6] = room_max.z;
	cube_info[0][7] = 5;  cube_info[0][8] = 0;
	
	num_cube++;
	cout << "Room dimensions: " << room_min.x << " "<< room_max.x << " " << room_min.y << " " << room_max.y << " " << room_min.z << " " << room_max.z << endl;

	// Reconstruction of object models in the room
	for (int i = 0; i < region_counter; i++)
	{
		if (region_info.at<int>(i,0) == 0 || region_info.at<int>(i, 1) == 2)		//invalid region or room rayout
			continue;

		generate_new_cuboid(i);		//Reconstruction with eroded regions

		for (int j = 1; j < 7; j++)
			new_cube_info[j] = floor((new_cube_info[j] / cube_resolusion) + 0.5) * cube_resolusion; // quantization to the cube-resolution

		for (int j = 0; j < 9; j++)
			cube_info[num_cube][j] = new_cube_info[j];

		num_cube++;
	}

	cout << "Number of total objects: " << num_cube << endl;

	//Memory for final data
	cube_valid_face = new int*[MAX_OBJECTS];
	cube_volume = new float[MAX_OBJECTS];

	for (int i = 0; i<MAX_OBJECTS; i++)
	{
		cube_valid_face[i] = new int[7];		//[0] empty [1]Min x [2]Max_x [3]Min y [4]max y [5]min z [6] max z
	}

	for (int i = 0; i<num_cube + 1; i++)
		for (int j = 0; j<7; j++)
			cube_valid_face[i][j] = 1;

	//Cube range refinement to fit into the room boundary
	cube_boundary_refinement(num_cube);
	cube_extension_to_ground(num_cube);


	for (int i = 0; i<num_cube; i++)
		cube_volume[i] = (cube_info[i][2] - cube_info[i][1])*(cube_info[i][4] - cube_info[i][3])*(cube_info[i][6] - cube_info[i][5]);

	//if (CUBE_REFINEMENT)		//adjust the range of cubes so that they don't cross each other
	//{
	//cube_range_refinement(num_cube);  // Not implemented yet
	//}

	remove_overlapping_faces(num_cube);

	//Mesh generation
	cout << "Cuboid info generation......" << endl;

	new_max_vertice = img_size.height * img_size.width / 8;
	new_max_face = new_max_vertice * 3;
	new_vertex = new CvPoint3D32f[new_max_vertice];
	vertex_visibility = new int[new_max_vertice];
	new_color = new CvPoint3D32f[new_max_vertice];
	new_alpha = new uchar[new_max_vertice];
	new_uv = new CvPoint2D32f[new_max_face];
	new_face = cvCreateMat(new_max_face, 6, CV_32S);
	new_rectangle = cvCreateMat(new_max_face, 8, CV_32S);
	new_valid_face = new int[new_max_face];

	for (int i = 0; i<new_max_vertice; i++)
		new_valid_face[i] = 1;

	for (int i = 0; i<new_max_vertice; i++)
		new_alpha[i] = 255;

	for (int i = 0; i<OBJ_CLASS_NUMBER; i++)
		obj_counter[i] = 0;

	int		obj_index, mat_index, mat_label;

	//Room boundary generation

	int no_object = 0;
	for (int i = 0; i < num_cube; i++)
	{
		if (cube_info[i][0] == 2)
		{
			obj_index = 5;  //Floor but other faces will have walls and ceiling in the final output

			obj_counter[obj_index]++;
			obj_list[no_object][0] = obj_index;
			obj_list[no_object][1] = obj_counter[obj_index];
			obj_list[no_object][2] = new_total_f;		//start rectagle 
			cube_rectangle_generation_boundary(cube_info, i);
			obj_list[no_object][3] = new_total_f - 1;	//end rectagle 
			int		mat_index = cube_info[no_object][8];
			int		mat_label;
			mat_list[no_object][0] = obj_index;

			mat_index = mat_index / 2;	//material[0] (none) doesn't exist

			for (int j = 1; j<12; j++)
			{
				mat_label = mat_index % 2;
				mat_list[no_object][j] = mat_label;
				mat_index = mat_index / 2;
			}

			no_object++;
		}
	}


	//Object cuboids generation
	for (int i = 0; i<num_cube; i++)
	{
		if (region_info.at<int>(i, 0) == 0 || region_info.at<int>(i, 1) == 0 || cube_info[i][0] != 1)
			continue;

		int		obj_index = cube_info[i][7];
		obj_counter[obj_index]++;
		obj_list[no_object][0] = obj_index;
		obj_list[no_object][1] = obj_counter[obj_index];
		obj_list[no_object][2] = new_total_f;		//start rectagle 

		cube_rectangle_generation(cube_info, i);
		
		obj_list[no_object][3] = new_total_f - 1;	//end rectagle 

		int		mat_index = cube_info[i][8];
		int		mat_label;
		mat_list[no_object][0] = obj_index;

		mat_index = mat_index / 2;	//material[0] (none) doesn't exist

		for (int j = 1; j<12; j++)
		{
			mat_label = mat_index % 2;
			mat_list[no_object][j] = mat_label;
			mat_index = mat_index / 2;
		}

		no_object++;
	}

	end_time = clock();
	cout << "Time for Cuboid reconstruction: " << ((double)(end_time - start_time) / CLOCKS_PER_SEC) << "secs" << endl;


	start_time = clock();

	//JSON buffer
	jwOpen(JSON_buffer, JSON_buflen, JW_OBJECT, JW_PRETTY);		// start JSON root object
	char object_name[512], object_full_name[512];
	no_object = 0;
	int err;

	for (int i = 0; i<num_cube; i++)	//For room bounday
	{
		if (cube_info[i][0] == 2)
		{
			//floor
			jwObj_object("Floor_1");					// object in the root object
			jwObj_string("Category", "Floor");
			jwObj_string("Material", "Wood");
			jwObj_string("Material", "Flat");
			jwObj_double("Centre_x", (cube_info[i][2] + cube_info[i][1]) / 2.0);
			jwObj_double("Centre_y", (cube_info[i][4] + cube_info[i][3]) / 2.0);
			jwObj_double("Centre_z", (cube_info[i][5]));
			jwObj_double("Length_x", cube_info[i][2] - cube_info[i][1]);
			jwObj_double("Length_y", cube_info[i][4] - cube_info[i][3]);
			jwObj_double("Length_z", 0.0);
			jwObj_double("absorption", 0.05);
			jwObj_double("scattering", 0.02);
			jwEnd();

			//ceiling
			jwObj_object("Ceiling_1");					// object in the root object
			jwObj_string("Category", "Ceiling");
			jwObj_string("Material", "Painted");
			jwObj_double("Centre_x", (cube_info[i][2] + cube_info[i][1]) / 2.0);
			jwObj_double("Centre_y", (cube_info[i][4] + cube_info[i][3]) / 2.0);
			jwObj_double("Centre_z", (cube_info[i][6]));
			jwObj_double("Length_x", cube_info[i][2] - cube_info[i][1]);
			jwObj_double("Length_y", cube_info[i][4] - cube_info[i][3]);
			jwObj_double("Length_z", 0.0);
			jwObj_double("absorption", 0.05);
			jwObj_double("scattering", 0.02);
			jwEnd();

			//Wall - at y_min
			jwObj_object("Wall_1");					// object in the root object
			jwObj_string("Category", "Wall");
			jwObj_string("Material", "Painted");
			jwObj_string("Material", "Wood");
			jwObj_string("Material", "Flat");
			jwObj_double("Centre_x", (cube_info[i][2] + cube_info[i][1]) / 2.0);
			jwObj_double("Centre_y", cube_info[i][4]);
			jwObj_double("Centre_z", (cube_info[i][6] + cube_info[i][5]) / 2.0);
			jwObj_double("Length_x", cube_info[i][2] - cube_info[i][1]);
			jwObj_double("Length_y", 0.0);
			jwObj_double("Length_z", cube_info[i][6] - cube_info[i][5]);
			jwObj_double("absorption", 0.05);
			jwObj_double("scattering", 0.02);
			jwEnd();

			//Wall - at x_max
			jwObj_object("Wall_2");					// object in the root object
			jwObj_string("Category", "Wall");
			jwObj_string("Material", "Painted");
			jwObj_string("Material", "Wood");
			jwObj_string("Material", "Flat");
			jwObj_double("Centre_x", cube_info[i][2]);
			jwObj_double("Centre_y", (cube_info[i][4] + cube_info[i][3]) / 2.0);
			jwObj_double("Centre_z", (cube_info[i][6] + cube_info[i][5]) / 2.0);
			jwObj_double("Length_x", 0.0);
			jwObj_double("Length_y", cube_info[i][4] - cube_info[i][3]);
			jwObj_double("Length_z", cube_info[i][6] - cube_info[i][5]);
			jwObj_double("absorption", 0.05);
			jwObj_double("scattering", 0.02);
			jwEnd();

			//Wall - at y_max
			jwObj_object("Wall_3");					// object in the root object
			jwObj_string("Category", "Wall");
			jwObj_string("Material", "Painted");
			jwObj_string("Material", "Wood");
			jwObj_string("Material", "Flat");
			jwObj_double("Centre_x", (cube_info[i][2] + cube_info[i][1]) / 2.0);
			jwObj_double("Centre_y", cube_info[i][6]);
			jwObj_double("Centre_z", (cube_info[i][6] + cube_info[i][5]) / 2.0);
			jwObj_double("Length_x", cube_info[i][2] - cube_info[i][1]);
			jwObj_double("Length_y", 0.0);
			jwObj_double("Length_z", cube_info[i][6] - cube_info[i][5]);
			jwObj_double("absorption", 0.05);
			jwObj_double("scattering", 0.02);
			jwEnd();

			//Wall - at x_min
			jwObj_object("Wall_4");					// object in the root object
			jwObj_string("Category", "Wall");
			jwObj_string("Material", "Painted");
			jwObj_string("Material", "Wood");
			jwObj_string("Material", "Flat");
			jwObj_double("Centre_x", cube_info[i][1]);
			jwObj_double("Centre_y", (cube_info[i][4] + cube_info[i][3]) / 2.0);
			jwObj_double("Centre_z", (cube_info[i][6] + cube_info[i][5]) / 2.0);
			jwObj_double("Length_x", 0.0);
			jwObj_double("Length_y", cube_info[i][4] - cube_info[i][3]);
			jwObj_double("Length_z", cube_info[i][6] - cube_info[i][5]);
			jwObj_double("absorption", 0.05);
			jwObj_double("scattering", 0.02);
			jwEnd();

			no_object++;
		}
	}


	for (int i = 0; i<num_cube; i++)
	{
		if (cube_info[i][0] != 1)
			continue;

		sprintf(object_full_name, "%s_%d", obj_name[obj_list[i][0]].c_str(), obj_list[i][1]);
		sprintf(object_name, "%s", obj_name[obj_list[i][0]].c_str());

		jwObj_object(object_full_name);					// object in the root object
		jwObj_string("Category", object_name);

		if (cube_info[i][12] == 0)
			jwObj_string("Material", "Unknown");
		else
		{
			for (int k = 1; k<12; k++)
			{
				if (mat_list[i][k] == 1)
				{
					sprintf(object_name, "%s", mat_name[k].c_str());
					jwObj_string("Material", object_name);
				}
			}
		}

		jwObj_double("Centre_x", (cube_info[i][2] + cube_info[i][1]) / 2.0);
		jwObj_double("Centre_y", (cube_info[i][4] + cube_info[i][3]) / 2.0);
		jwObj_double("Centre_z", (cube_info[i][6] + cube_info[i][5]) / 2.0);
		jwObj_double("Length_x", (cube_info[i][2] - cube_info[i][1]));
		jwObj_double("Length_y", (cube_info[i][4] - cube_info[i][3]));
		jwObj_double("Length_z", (cube_info[i][6] - cube_info[i][5]));
		jwObj_double("absorption", 0.05);
		jwObj_double("scattering", 0.02);
		jwEnd();

		no_object++;
	}


	err = jwClose();								// close and get error code
													//Saving data
	printf("Saving JSON file\n");
	FILE *output;
	if ((fopen_s(&output, (BASE_PATH + OUTPUT_PREFIXE + ".JSON").c_str(), "w")) != 0)
	{
		cout << "Could not open the JSON file to write" << endl;
		exit(0);
	}


	fprintf(output, "%s\n", JSON_buffer);
	fclose(output);


	printf("Saving cube info\n");

	FILE *output1;
	if ((fopen_s(&output1, (BASE_PATH + OUTPUT_PREFIXE + "_cuboid.txt").c_str(), "w")) != 0)
	{
		cout << "Could not open the cuboid info file to write" << endl;
		exit(0);
	}

	fprintf(output1, "%d %f\n", num_cube, cube_resolusion);

	for (int i = 0; i < num_cube; i++)
	{
		for (int j = 0; j < 9; j++)
			fprintf(output1, "%f ", cube_info[i][j]);
		fprintf(output1, "\n");
	}

	fclose(output1);


	if (OBJ_SAVE == 0)
		save_obj_rect_object(new_vertex, obj_list, no_object, new_rectangle, new_valid_face, new_total_v, new_total_f);
	else if (OBJ_SAVE == 1)
		save_obj_rect_object_colour(new_vertex, new_color, obj_list, no_object, new_rectangle, new_valid_face, new_total_v, new_total_f);
	else if (OBJ_SAVE == 2)
		save_obj_rect(new_vertex, new_uv, new_rectangle, new_valid_face, new_total_v, new_total_f);
	else if (OBJ_SAVE == 3)
		save_off_rect_object(new_vertex, new_color, new_alpha, new_rectangle, new_valid_face, new_total_v, new_total_f);

	end_time = clock();
	cout << "Time for Saving Output: " << ((double)(end_time - start_time) / CLOCKS_PER_SEC) << "secs" << endl;
	printf("Done!!!\n");
	return 0;

}

